# coding: utf-8

__all__ = ['get_news_headlines', 'get_news_story']

import pandas as pd
import numpy
import eikon.json_requests, eikon.Profile
from .tools import is_string_type, to_datetime, tz_replacer

News_Headlines_UDF_endpoint = "News_Headlines"
News_Story_UDF_endpoint = "News_Story"


def get_news_headlines(query='Topic:TOPALL and Language:LEN', count=10, date_from=None,
                       date_to=None, raw_output=False, debug=False):
    """
    Returns a list of news headlines

    Parameters
    ----------
    query: string, optional
        News headlines search criteria.
        The text can contain RIC codes, company names, country names and
        operators (AND, OR, NOT, IN, parentheses and quotes for explicit search…).

        Tip: Append 'R:' in front of RIC names to improve performance.

        Default: Top News written in English

    count: int, optional
        Max number of headlines retrieved.

        Value Range: [1-100].

        Default: 10

    date_from: string or datetime, optional
        Beginning of date range.

        String format is: '%Y-%m-%dT%H:%M:%S'. e.g. '2016-01-20T15:04:05'.

    date_to: string or datetime, optional
        End of date range.

        String format is: '%Y-%m-%dT%H:%M:%S'. e.g. '2016-01-20T15:04:05'.

    raw_output: boolean, optional
        Set this parameter to True to get the data in json format
        if set to False, the function will return a data frame.

        Default: False

    debug: bool, optional
        When set to True, the json request and response are printed.

        Default: False

    Returns
    -------
        pandas.DataFrame
            Returns a DataFrame of news headlines with the following columns:

            - Index               : Timestamp of the publication time
            - version_created     : Date of the latest update on the news
            - text                : Text of the Headline
            - story_id            : Identifier to be used to retrieve the full story using the get_news_story function
            - source_code         : Second news identifier

    Raises
    ----------
        Exception
            If http request fails or if server returns an error.

        AttributeError
            If a parameter type is wrong.

    Examples
    --------
    >>> import eikon as ek
    >>> ek.set_app_key('set your app id here')
    >>> headlines = ek.get_news_headlines("R:MSFT.O", 2)
    >>> headlines
    versionCreated                                   text
    2016-04-13 18:28:57.000 2.016-04-13 18:28:59.001 RBC Applies Blockchain as a Loyalty Boost...
    2016-04-13 17:28:21.577 2016-04-13 17:28:21.671 UPDATE 2-Long-stalled email privacy bill  ...
                               storyId
    2016-04-13 18:28:57.000    urn:newsml:reuters.com:20160413:nNRA1uxh03:1
    2016-04-13 17:28:21.577    urn:newsml:reuters.com:20160413:nL2N17G16Q:2

    >>> headlines = ek.get_news_headlines("R:MSFT.O IN FRANCE")
    >>> headlines = ek.get_news_headlines("R:MSFT.O IN FRANCE IN ENGLISH", count=5)
    >>> headlines = ek.get_news_headlines("OBA* OR CLINTON IN ENGLISH", count=5)
    """
    logger = eikon.Profile.get_profile().logger

    # check parameters type and values
    if not is_string_type(query):
        error_msg = 'query must be a string'
        logger.error(error_msg)
        raise ValueError(error_msg)

    if type(count) is not int:
        error_msg = 'count must be an integer'
        logger.error(error_msg)
        raise ValueError(error_msg)
    elif count < 0:
        error_msg = 'count must be equal or greater than 0'
        logger.error(error_msg)
        raise ValueError(error_msg)

    # build the payload
    app_key = eikon.get_app_key()
    payload = {'number': str(count), 'query': query, 'productName': app_key, 'attributionCode': ''}

    if date_from is not None:
        payload.update({'dateFrom': to_datetime(date_from).isoformat()})

    if date_to is not None:
        payload.update({'dateTo': to_datetime(date_to).isoformat()})

    result = eikon.json_requests.send_json_request(News_Headlines_UDF_endpoint, payload, debug=debug)

    if raw_output:
        return result
    else:
        return get_data_frame(result)


def get_data_frame(json_data):
    Headline_Selected_Fields = ['versionCreated', 'text', 'storyId', 'sourceCode']
    Headline_All_Fields_ = ['text', 'storyId', 'bodyType', 'displayDirection', 'documentType',
                             'isAlert', 'language', 'permIDs', 'products', 'rcs', 'reportCode', 'sourceCode',
                             'sourceName', 'versionCreated']
 
    json_headlines_array = json_data['headlines']
    first_created = [tz_replacer(headline['firstCreated']) for headline in json_headlines_array]
    headlines = [[headline[field] for field in Headline_Selected_Fields]
                 for headline in json_headlines_array]
    if len(headlines):
        headlines_dataframe = pd.DataFrame(headlines, numpy.array(first_created, dtype='datetime64'), Headline_Selected_Fields)
        headlines_dataframe['versionCreated'] = headlines_dataframe.versionCreated.apply(pd.to_datetime)
        headlines_dataframe = headlines_dataframe.convert_dtypes()
    else:
        headlines_dataframe = pd.DataFrame([], numpy.array(first_created, dtype='datetime64'), Headline_Selected_Fields)
                                           
    return headlines_dataframe


def get_news_story(story_id, raw_output=False, debug=False):
    """
    Return a single news story corresponding to the identifier provided in story_id

    Parameters
    ----------
    story_id: string
        The story id is a field you will find in every headline you retrieved with the function get_news_headlines.

    raw_output: boolean
        Set this parameter to True to get the data in json format
        if set to False, the function will return returns the story content.

        The default value is False.

    debug: boolean, optional
        When set to True, the json request and response are printed.
        Default: False

    Raises
    ------
        Exception
            If http request fails or if Refinitiv Services return an error.

        ValueError
            If a parameter type or value is wrong.

    Examples
    --------
    >>> import eikon as ek
    >>> ek.set_app_key('set your app key here')
    >>> headlines = ek.get_news_headlines('IBM')
    >>> for index, headline_row in headlines.iterrows():
            story = ek.get_news_story(headline_row['storyId'])
            print (story)
    """

    app_key = eikon.get_app_key()
    payload = {'attributionCode': '', 'productName': app_key, 'silent': True, 'storyId': story_id}
    json_data = eikon.json_requests.send_json_request(News_Story_UDF_endpoint, payload, debug=debug)

    if raw_output:
        return json_data
    else:
        return json_data['story']['storyHtml']





# coding: utf-8

__all__ = ['TR_Field', 'get_data']

import eikon.json_requests
import pandas as pd
import numpy
from .tools import get_json_value,is_string_type, check_for_string_or_list_of_strings, \
    check_for_string, build_dictionary, build_list, build_list_with_params


DataGrid_UDF_endpoint = 'DataGrid'
DataGridAsync_UDF_endpoint = 'DataGrid_StandardAsync'


def TR_Field(field_name, params=None, sort_dir=None, sort_priority=None):
    """
    This is a helper function to build the field for the get_data function.

    Parameters
    ----------
    field_name: string
        Field name to request. You can find the list in Data Item Browser.

    params: dict
        Dictionary containing the parameters for the field passed in the argument field_name

    sort_dir: string
        Indicate the sort direction. Possible values are 'asc' or 'desc'. The default value is 'asc'

    sort_priority: integer
        Gives a priority to the field for the sorting. The highest priority is 0 (zero). the default value is None

    Return
    ------
    Returns a dictionary that can directly passed to get_data.

    Example
    -------
    TR_Field('tr.revenue')
    TR_Field('tr.open','asc',1)
    TR_Field('TR.GrossProfit',{'Scale': 6, 'Curn': 'EUR'},'asc',0)

    """
    logger = eikon.Profile.get_profile().logger

    if params is not None and type(params) != dict:
        logger.error('TR_Field error: The argument params must be a dictionary')
        raise ValueError('TR_Field error: The argument params must be a dictionary')
    
    if type(params) == dict and not bool(params):
        error_msg = 'TR_Field error: The argument params must be a non empty dictionary or set to None (default value if not set)'
        logger.error(error_msg)
        raise ValueError(error_msg)

    field = {field_name:{}}
    if params: field[field_name]['params'] = params

    if sort_dir is not None:
        if is_string_type(sort_dir) and sort_dir in ['asc','desc']:
            field[field_name]['sort_dir'] = sort_dir
        else:
            error_msg = 'TR_Field error: The argument sort_dir must be a string ("asc" or "desc")'
            logger.error(error_msg)
            raise ValueError(error_msg)

    if sort_priority is not None:
        if type(sort_priority)is not int:
            error_msg = 'TR_Field error: The argument sort_priority must be a integer'
            logger.error(error_msg)
            raise ValueError(error_msg)
        field[field_name]['sort_priority'] = sort_priority
    return field


def get_data(instruments, fields, parameters=None, field_name=False, raw_output=False, debug=False):
    """
    Returns a pandas.DataFrame with fields in columns and instruments as row index

    Parameters
    ----------
    instruments: string or list
        Single instrument or list of instruments to request.

    fields: string, dictionary or list of strings and/or dictionaries.
        List of fields to request.

        Examples:

        - 'TR.PriceClose'
        - {'TR.GrossProfit': { 'params':{ 'Scale': 6, 'Curn': 'EUR' }}
        - {'TR.GrossProfit': { 'params':{ 'Scale': 6, 'Curn': 'EUR' },sort_dir:'desc'}
        - ['TR.PriceClose','TR.PriceOpen']
        - [{'TR.PriceClose':  {'sort_dir':asc,sort_priority:1}},{'TR.PriceOpen':  {'sort_dir':asc,sort_priority:0}}
        
        You can use the function TR_Field to build the fields:

        >>> fields = [ek.TR_Field('tr.revenue'),ek.TR_Field('tr.open','asc',1),ek.TR_Field('TR.GrossProfit',{'Scale': 6, 'Curn': 'EUR'},'asc',0)]
        >>> data, err = ek.get_data(["IBM","MSFT.O"],fields)
       
        Tips:
        You can launch the Data Item Browser to discover fields and parameters,
        or copy field names and parameters from TR Eikon - MS Office formulas

    parameters: string or dictionary, optional
        Single global parameter key=value or dictionary of global parameters to request.

        Default: None

    field_name: boolean, optional
        Define if column headers are filled with field name or display names.

        If True value, field names will ube used as column headers. Otherwise, the full display name will be used.

        Default: False

    raw_output: boolean, optional
        By default the output is a pandas.DataFrame.

        Set raw_output=True to get data in Json format.

        Default: False

    debug: boolean, optional
        When set to True, the json request and response are printed. Default value is False

    Returns
    -------
        pandas.DataFrame
            Returns pandas.DataFrame with fields in columns and instruments as row index

        errors
            Returns a list of errors

    Raises
    ----------
        Exception
            If http request fails or if server returns an error.

        ValueError
            If a parameter type or value is wrong.

    Examples
    --------
    >>> import eikon as ek
    >>> ek.set_app_key('set your app key here')
    >>> data, err = ek.get_data(["IBM", "GOOG.O", "MSFT.O"], ["TR.PriceClose", "TR.Volume", "TR.PriceLow"])
    >>> data, err = ek.get_data("IBM", ['TR.Employees', {'TR.GrossProfit':{'params':{'Scale': 6, 'Curn': 'EUR'},'sort_dir':'asc'}}])
    >>> fields = [ek.TR_Field('tr.revenue'),ek.TR_Field('tr.open',None,'asc',1),ek.TR_Field('TR.GrossProfit',{'Scale': 6, 'Curn': 'EUR'},'asc',0)]
    >>> data, err = ek.get_data(["IBM","MSFT.O"],fields)
    """
    logger = eikon.Profile.get_profile().logger

    check_for_string_or_list_of_strings(instruments, 'instruments')
    instruments = build_list(instruments, 'instruments')
    instruments = [value.upper() if value.islower() else value for value in instruments]

    if parameters:
        parameters = build_dictionary(parameters, 'parameters')

    fields = parse_fields(fields)
    fields_for_request = []
    for f in fields:
        keys =  list(f.keys())
        if len(keys) != 1:
            with 'get_data error: The field dictionary should contain a single key which is the field name' as msg:
                logger.error(msg)
                raise ValueError(msg)
        name = list(f.keys())[0]
        field_info = f[name]
        if type(field_info) != dict:
            with 'get_data error: The parameters for the file {} should be passed in a dict'.format(name) as error_msg:
                logger.error(error_msg)
                raise ValueError(error_msg)

        field = {'name':name}
        if 'sort_dir' in list(field_info.keys()): field['sort'] = field_info['sort_dir']
        if 'sort_priority' in list(field_info.keys()): field['sortPriority'] = field_info['sort_priority']
        if 'params' in list(field_info.keys()): field['parameters'] = field_info['params']
        fields_for_request.append (field)
     
    payload = {'instruments': instruments,'fields': fields_for_request}
    if parameters: payload.update({'parameters': parameters})

    _endpoint = DataGridAsync_UDF_endpoint

    if _endpoint == DataGridAsync_UDF_endpoint:
        payload = {'requests': [payload]}

    result = eikon.json_requests.send_json_request(_endpoint, payload, debug=debug)

    if result.get('responses'):
        result = result['responses'][0]

    if raw_output:
        return result

    return get_data_frame(result, field_name)


def parse_fields(fields):
    if is_string_type(fields):
        return [{fields: {}}]

    logger = eikon.Profile.get_profile().logger
    if type(fields) == dict:
        if len(fields) is 0:
            with 'get_data error: fields list must not be empty' as error_msg:
                logger.error(error_msg)
                raise ValueError(error_msg)
        return [fields]
    field_list = []
    if type(fields) == list:
        if len(fields) is 0:
            with 'get_data error: fields list must not be empty' as error_msg:
                logger.error(error_msg)
                raise ValueError(error_msg)
        for f in fields:
             if is_string_type(f):
                 field_list.append({f:{}})
             elif type(f) == dict:
                 field_list.append(f)
             else:
                 error_msg = 'get_data error: the fields should be of type string or dictionary'
                 eikon.Profile.get_profile().logger.error(error_msg)
                 raise ValueError(error_msg)
        return field_list

    error_msg = 'get_data error: the field parameter should be a string, a dictionary , or a list of strings|dictionaries'
    eikon.Profile.get_profile().logger.error(error_msg)
    raise ValueError(error_msg)


def get_data_value(value):
    if is_string_type(value):
        return value
    elif value is dict:
        return value['value']
    else:
        return value


def get_data_frame(data_dict, field_name=False):
    if field_name:
        headers = [header.get('field', header.get('displayName')) for header in data_dict['headers'][0]]
    else:
        headers = [header['displayName'] for header in data_dict['headers'][0]]
    data = numpy.array([[get_data_value(value) for value in row] for row in data_dict['data']])
    if len(data):
        df = pd.DataFrame(data, columns=headers)
    else:
        df = pd.DataFrame([], columns=headers)
    df = df.apply(pd.to_numeric, errors='ignore')
    if len(data):
        df = df.convert_dtypes()
    errors = get_json_value(data_dict, 'error')
    return df, errors



# coding: utf-8

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import re


# NoseTestCommand allow to launch nosetest with the command 'python setup.py test'
class NoseTestCommand(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # Run nose ensuring that argv simulates running nosetests directly
        import nose
        nose.run_exit(argv=['nosetests'])

try:
    from pypandoc import convert_file
    read_md = lambda f: convert_file(f, 'rst')
    description_content_type = None
except ImportError:
    print('##########################################################################')
    print("## warning: pypandoc module not found, could not convert Markdown to RST #")
    print('##########################################################################')
    print("")
    read_md = lambda f: open(f, 'r').read()
    description_content_type="text/markdown"


module_file = open("eikon/__init__.py").read()
metadata = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", module_file))

setup(name='eikon',
      version=metadata['version'],
      description='Python package for retrieving Eikon data.',
      long_description=read_md('README.md'),
      long_description_content_type=description_content_type,
      url='https://developers.refinitiv.com/eikon-apis/eikon-data-api',
      author='REFINITIV',
      author_email='',
      license='Apache 2.0',
      data_files=[("", ["LICENSE.md"])],
      packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
      use_2to3=True,
      zip_safe=False,
      install_requires=['requests>=2.22.0',
                        'requests-async>=0.6.2',
                        'nest_asyncio>=1.3.0',
                        'datetime',
                        'pandas>=1.0.0',
                        'numpy>=1.11.0',
                        'appdirs>=1.4.3',
                        'python-dateutil',
                        'websocket-client',
                        'deprecation'],
      test_suite='nose.collector',
      tests_require=['nose', 'mock', 'lettuce'],
      cmdclass={'test': NoseTestCommand}
)


# coding: utf-8

__all__ = ['Session', 'DacsParams']

import os
import logging
import socket
import requests_async as requests
import asyncio
import nest_asyncio
import time
from logging.handlers import RotatingFileHandler
from enum import Enum, unique
from datetime import datetime
from threading import Lock, Event, current_thread
from .stream_connection import StreamConnection
from ..eikonError import EikonError


# Load nest_asyncio to allow multiple calls to run_until_complete available
nest_asyncio.apply()


class DacsParams(object):

    def __init__(self, *args, **kwargs):
        self.dacs_username = kwargs.get("dacs_user_name", "user")
        self.dacs_application_id = kwargs.get("dacs_application_id", "256")
        self.dacs_position = kwargs.get("dacs_position")
        if self.dacs_position in [None, '']:
            try:
                position_host = socket.gethostname()
                self.dacs_position = "{}/{}".format(socket.gethostbyname(position_host), position_host)
            except socket.gaierror:
                self.dacs_position = "127.0.0.1/net"
        self.authentication_token = kwargs.get("authentication_token")


class Session(object):

    @unique
    class State(Enum):
        """
        Define the state of the session.
            Closed : The session is closed and ready to be opened.
            Pending : the session is in a pending state.
                Upon success, the session will move into an open state, otherwise will be closed.
            Open : The session is opened and ready for use.
        """
        Closed = 1
        Pending = 2
        Open = 3

    @classmethod
    def _state_msg(cls, state):
        if isinstance(state, Session.State):
            if state == Session.State.Opened:
                return "Session is Opened"
            if state == Session.State.Closed:
                return "Session is Closed"
            if state == Session.State.Pending:
                return "Session is Pending"
        return "Session is in an unknown state"  # Should not process this code path

    @unique
    class EventCode(Enum):
        """
        Each session can report different status events during it's lifecycle.
            StreamPending : Denotes the connection to the stream service within the session is pending.
            StreamConnected : Denotes the connection to the stream service has been successfully established.
            StreamDisconnected : Denotes the connection to the stream service is not established.
            SessionAuthenticationSuccess : Denotes the session has successfully authenticated this client.
            SessionAuthenticationFailed : Denotes the session has failed to authenticate this client.
            StreamAuthenticationSuccess: Denotes the stream has successfully authenticated this client.
            StreamAuthenticationFailed: Denotes the stream has failed to authenticate this client.
            TokenRefreshFailed  : The request to refresh the access token for the session has failed.
            DataRequestOk : The request for content from the sessions data services has completed successfully.
            DataRequestFailed : The request for content from the sessions data services has failed.
        """
        StreamPending = 1
        StreamConnected = 2
        StreamDisconnected = 3
        SessionAuthenticationSuccess = 4
        SessionAuthenticationFailed = 5
        StreamAuthenticationSuccess = 6
        StreamAuthenticationFailed = 7
        TokenRefreshFailed = 8
        DataRequestOk = 9
        DataRequestFailed = 10

    LOGGER_NAME = "pyeikon"

    class Params(object):
        def __init__(self, app_key=None, on_event=None, on_state=None, **kwargs):
            self._app_key = app_key
            self._on_event_cb = on_event
            self._on_state_cb = on_state
            self._dacs_params = DacsParams()

        def app_key(self, app_key):
            if app_key is None:
                raise AttributeError("app_key value can't be None")
            self._app_key = app_key
            return self

        def with_dacs_user_name(self, user):
            if user:
                self._dacs_params.dacs_username = user
            return self

        def with_dacs_application_id(self, application_id):
            if application_id:
                self._dacs_params.dacs_application_id = application_id
            return self

        def with_dacs_position(self, position):
            if position:
                self._dacs_params.dacs_position = position
            return self

        def on_state(self, on_state):
            self._on_state_cb = on_state
            return self

        def on_event(self, on_event):
            self._on_event_cb = on_event
            return self

    __all_sessions = {}
    __register_session_lock = Lock()
    __session_id_counter = 0

    @classmethod
    def register_session(cls, session):
        with cls.__register_session_lock:
            if not session:
                raise EikonError('Error', 'Try to register unavailable session')
            session_id = session.session_id
            if session_id in cls.__all_sessions:
                return
            session._session_id = cls.__session_id_counter
            cls.__session_id_counter += 1
            cls.__all_sessions[session._session_id] = session

    @classmethod
    def unregister_session(cls, session):
        with cls.__register_session_lock:
            if not session:
                raise EikonError('Error', 'Try to unregister unavailable session')
            session_id = session.session_id
            if session_id is None:
                raise EikonError('Error', 'Try to unregister unavailable session')
            if session_id not in cls.__all_sessions:
                raise EikonError('Error',
                                    'Try to unregister unknown session id {}'.format(session_id))
            cls.__all_sessions.pop(session_id)

    @classmethod
    def get_session(cls, session_id):
        """
        Returns the stream session singleton
        """
        if session_id not in cls.__all_sessions:
            raise EikonError('Error', 'Try to get unknown session id {}'.format(session_id))
        return cls.__all_sessions.get(session_id)

    def __init__(self, app_key, on_state=None, on_event=None,
                 token=None, dacs_user_name=None, dacs_position=None, dacs_application_id=None):
        from eikon.streaming_session.streaming_connection_config import StreamingConnectionConfiguration

        self._session_id = None
        self._lock_log = Lock()

        self._state = Session.State.Closed
        self._status = Session.EventCode.StreamDisconnected
        self._last_event_code = None
        self._last_event_message = None

        self._app_key = app_key
        self._on_event_cb = on_event
        self._on_state_cb = on_state
        self._access_token = token
        self._dacs_params = DacsParams()

        if dacs_user_name:
            self._dacs_params.dacs_username = dacs_user_name
        if dacs_position:
            self._dacs_params.dacs_position = dacs_position
        if dacs_application_id:
            self._dacs_params.dacs_application_id = dacs_application_id

        self._streaming_config = StreamingConnectionConfiguration()
        self._log_path = None
        self._log_level = logging.NOTSET

        logging.basicConfig(format=Session.FORMAT)
        logging.addLevelName(5, 'TRACE')
        self._logger = logging.getLogger(self.LOGGER_NAME)
        setattr(self._logger, 'trace', lambda *args: self._logger.log(5, *args))

        try:
            self._loop = asyncio.get_event_loop()
            self.log(1, f'Session loop was set to current event loop {self._loop}')
        except RuntimeError:
            self._loop = asyncio.new_event_loop()
            self.log(1, f'Session loop was set with a new event loop {self._loop}')
        self._streaming_session = None
        self._is_closing = False
        self._login_event = Event()
        self._login_event.clear()

        self.__lock_callback = Lock()
        self._http_session = requests.sessions.Session()
        self._base_url = 'https://api.edp.thomsonreuters.com'

        self._start_streaming_event = Event()
        self._stop_streaming_event = Event()
        self._stream_register_lock = Lock()
        self._all_stream_subscriptions = {}
        self._id_request = 0

        Session.register_session(self)

        if app_key is None:
            raise AttributeError("app_key value can't be None")

    def __del__(self):
        self.log(1, f'Delete a Session')
        Session.unregister_session(self)
        handlers = self._logger.handlers[:]
        for handler in handlers:
            handler.close()
            self._logger.removeHandler(handler)

    def __delete__(self, instance):
        self.log(1, f'Delete the Session instance {instance}')

    def _set_proxy(self, http, https):
        self._http_session.proxies = {"http": http, "https": https}

    def get_open_state(self):
        """
        Returns the session state.
        """
        return self._state

    def get_last_event_code(self):
        """
        Returns the last session event code.
        """
        return self._last_event_code

    def get_last_event_message(self):
        """
        Returns the last event message.
        """
        return self._last_event_message

    @property
    def app_key(self):
        """
        Returns the application id.
        """
        return self._app_key

    @app_key.setter
    def app_key(self, app_key):
        """
        Set the application key.
        """
        from eikon.tools import is_string_type

        if app_key is None:
            return
        if not is_string_type(app_key):
            raise AttributeError('application key must be a string')

        self._app_key = app_key

    @property
    def session_id(self):
        return self._session_id

    def logger(self):
        return self._logger

    ######################################
    # methods to manage log              #
    ######################################
    TRACE = 5
    MAX_LOG_SIZE = 10000000
    FORMAT = '%(asctime)-15s P[%(process)d] [%(threadName)s %(thread)s] %(message)s'
    #FORMAT = '%(asctime) - %(message)s'
    #FORMAT = '%(asctime),%(msecs)d %(levelname)-8s] %(message)s'
    # datefmt = '%Y-%m-%d:%H:%M:%S'
    # FORMAT = '[%(asctime)s] %(levelname)s - %(message)s'

    def set_log_path(self, log_path):
        """
        Set the path where log files will be created.

        Parameters
        ----------
        log_path : path directory
        Default: current directory (beside *.py running file)
        Return True if log_path exists and is writable
        """
        if os.access(log_path, os.W_OK):
            self._log_path = log_path
            return True
        return False

    def set_log_level(self, log_level):
        """
        Set the log level.
        By default, logs are disabled.

        Parameters
        ----------
        level : int
            Possible values from logging module :
            [CRITICAL, FATAL, ERROR, WARNING, WARN, INFO, DEBUG, NOTSET]
        """
        if log_level > logging.NOTSET:
            _formatter = logging.Formatter("[%(asctime)s;s] - [%(levelname)s] - [%(funcName)s] - %(message)s")
            _filename = "pyeikon.{}.log".format(datetime.now().strftime('%Y%m%d.%H-%M-%S'))

            if self._log_path is not None:
                if not os.path.isdir(self._log_path):
                    os.makedirs(self._log_path)
                _filename = os.path.join(self._log_path, _filename)

            _handler = RotatingFileHandler(_filename, mode='a',
                                           maxBytes=Session.MAX_LOG_SIZE,
                                           backupCount=10, encoding='utf-8')
            _handler.setFormatter(_formatter)
            self._logger.addHandler(_handler)

        self._logger.setLevel(log_level)
        self._log_level = log_level

        # set log_level to asyncio module
        logging.getLogger('asyncio').setLevel(log_level)

    def get_log_level(self):
        """
        Returns the log level
        """
        return self._logger.level

    def log(self, log_level, message):
        with self._lock_log:
            self._print(log_level, message)
            self._logger.log(log_level, message)

    def trace(self, message):
        self._logger.log(Session.TRACE, message)

    def _print(self, log_level, message):
        if False:
            print(f"{datetime.now()} - Thread {current_thread().ident} | {current_thread().name}\n{message}")

    ######################################
    # methods to open and close session  #
    ######################################
    def open(self):
        if self._state in [Session.State.Pending, Session.State.Open]:
            # session is already opened or is opening
            return self._state
        self._loop.run_until_complete(self.open_async())
        return self._state

    def close(self):
        if self._state == Session.State.Closed:
            return self._state

        self._state = Session.State.Closed
        self._stop_streaming()
        # Session.unregister_session(self)

        return self._state

    async def open_async(self):
        #Session.register_session(self)
        return self._state

    async def wait_for_streaming(self):
        await self._start_streaming()
        if self._status is Session.EventCode.StreamConnected:
            return True
        else:
            self.log(logging.DEBUG, "Streaming failed to start")
            return False

    async def _start_streaming(self):
        if self._status not in [Session.EventCode.StreamConnected,
                                Session.EventCode.StreamPending]:
            self._status = Session.EventCode.StreamPending

            self._start_streaming_event.clear()
            self._stop_streaming_event.clear()
            self._start_streaming_event = Event()
            self._stop_streaming_event = Event()
            self._login_event = Event()

            self._is_closing = False

            if self._streaming_session is None:
                _ws_name = "WebSocket {}".format(self.session_id)
                self.log(1, "Create StreamConnection...")
                self._streaming_session = StreamConnection(_ws_name, self,
                                                           self._start_streaming_event,
                                                           self._stop_streaming_event)
                self._streaming_session.daemon = True
                # Init web socket to support an open
                self._streaming_session.start()
                self.log(logging.DEBUG, "Streaming is started")
                time.sleep(0.2)

        self._start_streaming_event.set()

        if not self._login_event.is_set():
            self.log(1, "WAIT FOR LOGIN EVENT")
            self._login_event.wait()
            self.log(1, "RECEIVE LOGIN EVENT")
        else:
            self.log(1, "Session is logged, ")

        self._start_streaming_event.clear()
        return self._status

    def _send(self, msg):
        if self._streaming_session is not None:
            self._streaming_session.send(msg)

    def is_closing(self):
        return self._is_closing

    def _stop_streaming(self):
        # unblock any wait on login event
        self._is_closing = True
        if self._streaming_session:
            self._streaming_session.is_closing = True
        self.log(logging.DEBUG, f"Unlock login_event for streaming session {self._session_id} due to stop streaming call")
        if self._login_event:
            self._login_event.set()
            self._login_event.clear()
        if self._start_streaming_event:
            # unlock web socket on start_streaming event
            self._start_streaming_event.set()

        # Close web socket
        if self._streaming_session is not None:
            self._streaming_session.close()
            self._stop_streaming_event.wait()
            self.log(logging.INFO, f"Streaming session {self._session_id} was closed")
            del self._streaming_session
            self._streaming_session = None
        self._status = Session.EventCode.StreamDisconnected

    ##########################################################
    # methods for stream register / unregister               #
    ##########################################################
    def _get_new_id(self):
        self._id_request += 1
        return self._id_request

    def _register_stream(self, stream):
        with self._stream_register_lock:
            if stream is None:
                raise EikonError('Error', 'Try to register None subscription')

            if stream._stream_id in self._all_stream_subscriptions:
                raise EikonError('Error', f"Subscription {stream._stream_id} is already registered")
            if stream._stream_id is not None:
                raise EikonError('Error', f"Try to register again subscription {stream._stream_id}")
            stream._stream_id = self._get_new_id()
            self._all_stream_subscriptions[stream._stream_id] = stream

    def _unregister_stream(self, stream):
        with self._stream_register_lock:
            if not stream or not stream._stream_id:
                raise EikonError(-1, 'Try to unregister unavailable stream')

            if stream._stream_id not in self._all_stream_subscriptions:
                raise EikonError('Error',
                                 f"Try to unregister unknown stream {stream._stream_id} from session {self.session_id}")

            self._all_stream_subscriptions.pop(stream._stream_id)
            stream._stream_id = None

    def _get_stream(self, stream_id):
        with self._stream_register_lock:
            if stream_id is None:
                raise EikonError('Error', 'Try to retrieve undefined stream')
            if stream_id in self._all_stream_subscriptions:
                return self._all_stream_subscriptions[stream_id]
            return None

    ##########################################################
    # methods for session callbacks from streaming session   #
    ##########################################################
    def _on_open(self):
        with self.__lock_callback:
            self._state = Session.State.Pending
            pass

    def _on_close(self):
        with self.__lock_callback:
            self._state = Session.State.Closed
            pass

    def _on_state(self, state_code, state_text):
        with self.__lock_callback:
            if isinstance(state_code, Session.State):
                self._state = state_code
                if self._on_state_cb is not None:
                    self._on_state_cb(self, state_code, state_text)

    def _on_event(self, streaming_session_id, event_code, event_msg):
        with self.__lock_callback:
            if self._streaming_session:
                if streaming_session_id == self._streaming_session.streaming_session_id:
                    if isinstance(event_code, Session.EventCode):
                        if self._status != event_code:
                            self._status = event_code
                            if self._on_event_cb:
                                self._on_event_cb(self, event_code, event_msg)
                            # Unlock wait for login event if stream is disconnected
                            if event_code == Session.EventCode.StreamDisconnected:
                                self.log(logging.DEBUG,
                                         f"Unlock login_event for streaming session {self._streaming_session.streaming_session_id} due to disconnect event")
                                self._login_event.set()
                else:
                    # notification from another streaming session than current one
                    self.log(1, f'Received notification from another streaming session ({streaming_session_id}) than current one ({self._streaming_session.streaming_session_id})')
            else:
                self.log(1, f'Received notification for closed streaming session {streaming_session_id}')

    def process_on_close_event(self):
        self.close()

    ##############################################
    # methods for status reporting               #
    ##############################################
    @staticmethod
    def _report_session_status(self, session, session_status, event_msg):
        _callback = self._get_status_delegate(session_status)
        if _callback is not None:
            json_msg = self._define_results(session_status)[Session.CONTENT] = event_msg
            _callback(session, json_msg)

    def report_session_status(self, session, event_code, event_msg):
        # Report the session status event defined with the eventMsg to the appropriate delegate
        self._last_event_code = event_code
        self._last_event_message = event_msg
        _callback = self._get_status_delegate(event_code)
        if _callback is not None:
            _callback(session, event_code, event_msg)

    def _get_status_delegate(self, event_code):
        _cb = None

        if event_code in [Session.EventCode.SessionAuthenticationSuccess,
                          Session.EventCode.SessionAuthenticationFailed,
                          Session.EventCode.TokenRefreshFailed]:
            _cb = self._on_state_cb
        elif event_code not in [self.EventCode.DataRequestOk,
                                self.EventCode.StreamPending,
                                self.EventCode.StreamConnected,
                                self.EventCode.StreamDisconnected]:
            _cb = self._on_event_cb
        return _cb

    ############################
    # methods for HTTP request #
    ############################
    async def http_request_async(self, url: str, method=None, headers={},
                                 data=None, params=None, json=None, closure=None,
                                 auth=None, loop=None, **kwargs):
        if method is None:
            method = 'GET'

        if self._access_token is not None:
            headers["Authorization"] = "Bearer {}".format(self._access_token)

        if closure is not None:
            headers["Closure"] = closure

        headers.update({'x-tr-applicationid': self.app_key})

        _http_request = requests.Request(method, url, headers=headers, data=data, params=params, json=json, auth=auth,
                                         **kwargs)
        _prepared_request = _http_request.prepare()

        self.log(logging.DEBUG,
                 f'Request to {_prepared_request.url}\n   headers = {_prepared_request.headers}\n   params = {kwargs.get("params")}')

        try:
            async with requests.sessions.Session() as _http_session:
                _request_response = await _http_session.send(_prepared_request, **kwargs)
                self.log(1, f'HTTP request response {_request_response.status_code}: {_request_response.text}')
                return _request_response
        except Exception as e:
            self.log(1, f'HTTP request failed: {e!r}')

        return None

    def http_request(self, url: str, method=None, headers={}, data=None, params=None,
                     json=None, auth=None, loop=None, **kwargs):
        # Multiple calls to run_until_complete were allowed with nest_asyncio.apply()
        if loop is None:
            loop = self._loop
        response = loop.run_until_complete(self.http_request_async(url, method, headers, data,
                                                                   params, json, auth, **kwargs))
        return response



# coding: utf-8

__all__ = ['StreamConnection', 'StreamConnectionState']

import json
import websocket
import threading
import logging
from enum import Enum
from .stream import StreamState

import datetime


class StreamConnectionState(Enum):
    CLOSED = 0
    PENDING = 1
    OPEN = 2


class StreamConnection(threading.Thread):

    __all_streaming_session = {}
    __register_lock = threading.Lock()
    __streaming_session_id_counter = 0

    @classmethod
    def _get_new_streaming_session_id(cls):
        cls.__streaming_session_id_counter += 1
        return cls.__streaming_session_id_counter

    @classmethod
    def register_streaming_session(cls, streaming_session):
        with cls.__register_lock:
            if not streaming_session:
                raise EikonError('Error', 'Try to register unavailable streaming session')
            streaming_session_id = streaming_session.streaming_session_id
            if streaming_session_id in cls.__all_streaming_session:
                raise EikonError('Error',
                                 f'Try to register again existing streaming session id {streaming_session_id}')

            streaming_session._streaming_session_id = cls._get_new_streaming_session_id()
            cls.log(streaming_session, 1,
                    "Register streaming session {}".format(streaming_session._streaming_session_id))
            cls.__all_streaming_session[streaming_session._streaming_session_id] = streaming_session

    @classmethod
    def unregister_streaming_session(cls, streaming_session):
        with cls.__register_lock:
            if not streaming_session:
                raise EikonError('Error', 'Try to unregister unavailable streaming session')
            if streaming_session.streaming_session_id is None:
                raise EikonError('Error', 'Try to unregister unavailable streaming session id')
            if streaming_session.streaming_session_id not in cls.__all_streaming_session:
                raise EikonError('Error',
                                    'Try to unregister unknown streaming session id {}'
                                    .format(streaming_session.streaming_session_id))
            cls.log(streaming_session, 1,
                    f'Unregister streaming session {streaming_session._streaming_session_id}')
            cls.__all_streaming_session.pop(streaming_session.streaming_session_id)

    __all_subscriptions = {}
    __id_request = 1

    ###############################################################################################

    def __init__(self, thread_name, session, start_event, stop_event, *args, **kwargs):
        from eikon.streaming_session.session import Session

        if session is None:
            raise ValueError("StreamConnection is passed a null session")
        if session._streaming_config is None:
            raise ValueError("StreamConnection must have a WebsocketEndpoint")
        if session._streaming_config.url is None:
            raise ValueError("StreamConnection must have a WebsocketEndpoint")

        self._streaming_session_id = None
        self._streaming_config = session._streaming_config
        self._session = session
        self._on_state_cb = session._on_state
        self._on_event_cb = session._on_event
        self._start_streaming_event = start_event
        self._stop_streaming_event = stop_event
        self._is_closing = False

        self._ws_login_id = None
        self._websocket = None
        self._ws_connected = False
        self._ws_is_logged = False
        self._ws_lock = threading.Lock()

        self._logger = logging.getLogger(Session.LOGGER_NAME)
        self._state = StreamConnectionState.CLOSED

        threading.Thread.__init__(self, target=self.run, name=thread_name)
        StreamConnection.register_streaming_session(self)

    def __del__(self):
        self.log(1, 'StreamConnection {} is releasing'.format(self._streaming_session_id))
        if self._websocket:
            try:
                if self._websocket.keep_running:
                    # Close web socket
                    self.log(1, "Close websocket client {}".format(self._streaming_session_id))
                    self._websocket.close()
                    self._websocket.keep_running = False

            except Exception as e:
                self.log(1, f'Exception on close websocket attempt for main stream {self._streaming_session_id}: {e!r}')
                pass
        if self._streaming_session_id in StreamConnection.__all_streaming_session:
            self.log(1, "Unregister streaming session {}".format(self._streaming_session_id))
            StreamConnection.unregister_streaming_session(self)

    def username(self, user):
        self._streaming_config.username = user
        return self

    def position(self, position):
        self._streaming_config.position = position

    def application_id(self, app_id):
        self._streaming_config.application_id = app_id

    def auth_token(self, token):
        self._streaming_config.auth_token = token

    def connection_retry(self, retry_in_seconds):
        self._streaming_config.connection_retry = retry_in_seconds

    def log(self, log_level, message):
        if self._logger:
            self._logger.log(log_level, message)

    @property
    def streaming_session_id(self):
        return self._streaming_session_id

    @property
    def is_connected(self):
        return self._ws_connected

    @property
    def is_closing(self):
        return self._is_closing

    @is_closing.setter
    def is_closing(self, value):
        self._is_closing = value

    #############################################
    #  methods to open and close the websocket  #
    #############################################
    def run(self):

        while not self.is_closing:
            self.log(1, f"Streaming session {self._streaming_session_id} waits for start event")
            self._start_streaming_event.wait()
            if not self.is_closing:
                self.log(1, f"Streaming session {self._streaming_session_id} received start event, then open websocket.")

                self._websocket = websocket.WebSocketApp(self._streaming_config.url,
                                                         header=["User-Agent: Python"]+self._streaming_config._header,
                                                         on_message=self._ws_message,
                                                         on_error=self._ws_error,
                                                         on_close=self._ws_close,
                                                         subprotocols=["tr_json2"])
                self._websocket.on_open = self._ws_open
                self._websocket.id = self._streaming_session_id
                self._state = StreamConnectionState.PENDING
                self._websocket.run_forever()
                self._websocket = None
                self.log(1, "Websocket for streaming session {} was closed".format(self._streaming_session_id))
        self.log(1, "Streaming session {} will be closed".format(self._streaming_session_id))


    def close(self):
        if not self._is_stopped:

            # set notify flag to Flase to avoid any more on_event

            # Send Close message for the web socket
            close_json = {'Type': 'Close'}
            close_json['ID'] = self._ws_login_id
            close_json['Domain'] = 'Login'
            self.log(1, "Send Close message for main stream {} (login id {})".format(self._streaming_session_id,
                                                                                     self._ws_login_id))
            self.send(close_json)
            StreamConnection.unregister_streaming_session(self)
            if self._websocket and self._websocket.keep_running:
                # Close web socket
                self.log(1, "Close websocket client {}".format(self._streaming_session_id))
                self._websocket.close()
                # self._websocket.keep_running = False


    #############################################
    #  methods to send request to the websocket #
    #############################################
    def send(self, request):
        if self._websocket:
            self.log(1, "Send request: {}".format(request))
            self._send(request)

    ############################################
    # Methods for web socket callbacks         #
    ############################################
    def _ws_open(self, *args):
        from eikon.streaming_session.session import Session
        with self._ws_lock:
            result = "WebSocket for streaming session {} was opened to server: {}".format(self._streaming_session_id,
                                                                    self._streaming_config.url)
            self.log(1, result)
            if self._on_event_cb:
                self._on_event_cb(self._streaming_session_id,
                                  Session.EventCode.StreamConnected,
                                  result)
            self._ws_connected = True
            _login_json = self._streaming_config.login_message
            self._ws_login_id = self._session._get_new_id()
            _login_json['ID'] = self._ws_login_id
            self.send(_login_json)
        pass

    def _ws_error(self, error):
        from eikon.streaming_session.session import Session
        with self._ws_lock:
            err = "WebSocket error occurred for web socket client {} (login id {}) : {}".format(self._streaming_session_id,
                                                                                                self._ws_login_id,
                                                                                                error)
            self.log(1, err)
            self._ws_is_logged = False
            self._ws_connected = False
            self._ws_login_id = None
            if self._on_event_cb:
                self._on_event_cb(self._streaming_session_id,
                                  Session.EventCode.StreamDisconnected,
                                  err)
            self._state = StreamConnectionState.CLOSED
            self._stop_streaming_event.set()
            # Stop iteration on web socket connexion attempt in run() function
            self.is_closing = True
        pass

    def _ws_message(self, *args):
        with self._ws_lock:
            if self._on_event_cb:
                self._on_message(args[0])

    def _ws_close(self, *args):
        from eikon.streaming_session.session import Session
        with self._ws_lock:
            self._state = StreamConnectionState.CLOSED
            self.log(1, "Close notification from main stream {} (login id {})".format(self._streaming_session_id,
                                                                                      self._ws_login_id))
            self._ws_is_logged = False
            self._ws_connected = False
            self._ws_login_id = None
            if self._on_event_cb:
                self._on_event_cb(self._streaming_session_id,
                                  Session.EventCode.StreamDisconnected,
                                  "Connection to the WebSocket server [{}] is down".format(self._streaming_config.url))
            self._state = StreamConnectionState.CLOSED
            self._stop_streaming_event.set()

    ############################################
    # Send request method                      #
    ############################################
    def _send(self, msg):
        try:
            if self._ws_connected:
                self._websocket.send(json.dumps(msg))
            pass
        except websocket.WebSocketConnectionClosedException as e:
            self._logger.log(logging.ERROR, "WebSocketConnectionClosedException: {}".format(e))

    ###############################################
    # Parse methods for _on_message notifications #
    ###############################################
    def _on_message(self, message):
        """ Called when message is received from websocket"""
        message_json = json.loads(message)
        self.log(logging.DEBUG, 'Receive message from Web Socket')
        for singleMsg in message_json:
            self._process_message(singleMsg)

    def _process_message(self, message_json):
        """ Parse at high level and output JSON of message """
        from eikon.streaming_session.session import Session

        if self._session.is_closing():
            return

        message_type = message_json['Type']
        _id = message_json.get("ID")
        if _id == self._ws_login_id:
            self.log(Session.TRACE, f"Receive message for login {_id}: {message_json}")
        else:
            self.log(Session.TRACE, f"Receive message for stream {_id}: {message_json}")

        if message_type == "Refresh":
            if 'Domain' in message_json:
                message_domain = message_json['Domain']
                if message_domain == "Login":
                    self._process_login_response(message_json)
                    return
            self._process_refresh_message(message_json)
        elif message_type == 'Update':
            self._process_update_message(message_json)
        elif message_type == 'Status':
            if 'Domain' in message_json:
                message_domain = message_json['Domain']
                if message_domain == "Login":
                    self._process_login_response(message_json)
                    return
            self._process_status_message(message_json)
        elif message_type == 'Error':
            self._process_error_message(message_json)
        elif message_type == "Ping":
            self.log(logging.INFO, 'Receive ping from server ...')
            pong_json = {'Type': 'Pong'}
            self.send(pong_json)
            self.log(logging.INFO, '    ... send pong response')

    def _process_login_response(self, response):
        """ Parse login response message """
        from eikon.streaming_session.session import Session
        id = response.get("ID")
        if id != self._ws_login_id:
            self.log(logging.DEBUG, f'Received login response for id {id} different than login id {self._ws_login_id}')
        else:
            self.log(logging.DEBUG, f'Received login response for login id {id}')
        state = response.get("State")
        stream_status = state.get("Stream")
        data_status = state.get("Data")
        if stream_status == "Open" and data_status == "Ok":
            self._state = StreamConnectionState.OPEN
            self._ws_is_logged = True
            self.log(logging.INFO, "Login to websocket successful")
            self._on_event_cb(self._streaming_session_id,
                              Session.EventCode.StreamConnected,
                              state.get("Text"))
        else:
            self._state = StreamConnectionState.CLOSED
            self._on_event_cb(self._streaming_session_id,
                              Session.EventCode.StreamDisconnected,
                              "Login to websocket failed: {}".format(response))
        # Unblock all tasks that are waiting for Login response
        self._set_login_event()
        pass

    def _set_login_event(self):
        self.log(logging.DEBUG, f"Unlock login event due to login response")
        self._session._login_event.set()

    def _process_status_message(self, status):
        """ Parse status message """
        self.log(logging.INFO, 'Received status message:\n   {}'.format(status['State']))

        # EventCode value determines further interest
        _stream = str(status["State"]["Stream"]) if "State" in status and "Stream" in status["State"] else None
        # noInterest = (_stream == "Closed") or (_stream == "ClosedRecover");

        _id = status['ID']
        mp_subscription = self._session._get_stream(_id)
        if mp_subscription:
            mp_subscription._on_status(status)
            # _name = str(status["Key"]["Name"]) if "Key" in status and "Name" in status["Key"] else ""
            # if noInterest:
            #    mp_subscription._on_stream_state(StreamConnectionState.CLOSED)
        else:
            self.log(logging.WARNING, "Receive status message for unknown subscription id {}".format(_id))

    def _process_refresh_message(self, refresh):
        _id = refresh['ID']
        mp_subscription = self._session._get_stream(_id)

        # A completion occurs when "Complete" == true or the "Complete" field is absent
        _complete = str(refresh["Complete"]).lower() == "true" if refresh.get("Complete") else True

        # EventCode value determines further interest
        _stream = str(refresh["State"]["Stream"]) if refresh.get("State") and refresh["State"].get("Stream") else None
        _no_interest = _stream == "Closed" or _stream == "NonStreaming"

        if mp_subscription:
            mp_subscription._on_refresh(refresh)

            # A completion occurs when "Complete" == true or the "Complete" field is absent
            complete = True if refresh.get("Complete") is None else bool(refresh["Complete"]);
            if complete:
                mp_subscription._on_complete()
        else:
            self.log(logging.WARNING, "Receive refresh message for unknown subscription {}".format(_id))

    def _process_update_message(self, update):
        _id = update['ID']
        mp_subscription = self._session._get_stream(_id)
        if mp_subscription:
            mp_subscription._on_update(update)
        else:
            self.log(logging.WARNING, "Receive update message for unknown subscription {}".format(_id))

    def _process_error_message(self, error):
        _id = error['ID']
        mp_subscription = self._session._get_stream(_id)
        if mp_subscription:
            mp_subscription._on_error(error)
        elif _id == self._ws_login_id:
            self.log(logging.WARNING, "Receive error message for session {} : {}".format(_id, error))
            self._on_event_cb(self._streaming_session_id,
                              Session.EventCode.StreamDisconnected,
                              error)
        else:
            self.log(logging.WARNING, "Receive error message for unknown subscription {} : {}".format(_id, error))

    ##############################################
    # methods for refresh token                  #
    ##############################################
    def refresh_token(self, refresh_token):
        if self._ws_connected and self._ws_is_logged:
            refresh = {"ID": self._ws_login_id,
                       "Domain": "Login",
                       "Key": {
                            "NameType": "AuthnToken",
                            "Elements": {
                                "AuthenticationToken": refresh_token,
                                "ApplicationId": self._streaming_config.application_id,
                                "Position": self._streaming_config.position
                            }
                       },
                       "Refresh": False}
            self.send(refresh)

